/*
This script performs an automated training and validation run of a MemBraion neural net:

This is what the script does:
- Load and randomize the neural net
- Load a training lesson from an mbl file
- Load a validation lesson from an mbl file
- Train the neural net for an adjustable time using the training lesson
- During the training visualize the net error for both training and validation lesson (inter-changing)
- During the training visualize the net output in reaction to both training and validation 
  lesson for all output neurons (inter-changing) 
- After the training export the validation output of the net into a csv file. The export contains the 
  validation inputs, the validation outputs of the net and the original validation target output data.

The net also prints information to the Script Trace window during execution

Several parameters can be adjusted in the beginning of the script.
*/

// This script uses timers. Always copy the script 'SecondsTimer.as' into the sub directory 'SecondsTimer'.
#include "SecondsTimer\SecondsTimer.as"


// --------------------------------- Adjustable constants section begin --------------------------------------

string sNetName = "Net1.mbn";                    // Name of the neural net
string sTrainLessonName = "Train.mbl";           // Name of the training lesson file
string sValidateLessonName = "Validate.mbl";     // Name of the validation lesson file
string sResultFileName = "ValidationResult.csv"; // Name of the validation result file name
string sTargetColumnExtension = "(Target)";		 // Column name extension string for target output values

// The overall training time [seconds]
uint TRAIN_TIME = 10;

// Number of seconds the training lesson shall be displayed in the net error window.
uint SHOW_TRAIN_LESSON_TIME = 4;

// Number of seconds the validation lesson shall be displayed in the net error window.
uint SHOW_VALIDATE_LESSON_TIME = 10;

// Number of sesonds after which the monitored output neuron shall be toggled
uint TOGGLE_MONITORED_OUTPUT_INTERVAL = 2;

// --------------------------------- Adjustable constants section end --------------------------------------


// Program entry point
void main()
{
    OpenNet(sNetName);
    RandomizeNet();
    
    SetLessonCount(2);
    
    // Load validation lesson to lesson #2
    SelectLesson(2);
    LoadLesson(sValidateLessonName);
    
    // Load train lesson to lesson #1
    SelectLesson(1);
    LoadLesson(sTrainLessonName);
    
    // Start with training lesson as the net error lesson.
    SelectNetErrLesson(1);
    
    SelectTeacher("RPROP");
    
    ShowErrorViewer(true);
    ShowPatternErrorViewer(true);
    
    // Perform the training section of the script
    ProcessTraining();
    
    // Record the validation lesson to CSV
    RecordValidationLesson();
}


// Perform the training section of the script
void ProcessTraining()
{
    // Start the actual training
    StartTeaching();

	// Create the timer used to limit the overall training time
    SecondsTimer trainTimer(TRAIN_TIME);
    // Create the timer used to switch between training and validation lesson.
    SecondsTimer lessonTimer(SHOW_TRAIN_LESSON_TIME);
    // Create the timer used to toggle the monitored output neuron.
    SecondsTimer outputToggleTimer(TOGGLE_MONITORED_OUTPUT_INTERVAL);
        
    Trace("Display: Train Lesson\n");
    
    uint outputNeuron = 1;
    uint outputCount = GetOutputCount();
	    
    while (!trainTimer.IsElapsed() && IsTeaching())
    {
        if (lessonTimer.IsElapsed())
        {
            if (GetSelectedNetErrLesson() == 1)
            { // Switch net error lesson to be the validation lesson
                lessonTimer.Start(SHOW_VALIDATE_LESSON_TIME);
                Trace("Display: Validation Lesson\n");
                SelectNetErrLesson(2);
            }
            else
            { // Switch net error lesson to be the training lesson
                lessonTimer.Start(SHOW_TRAIN_LESSON_TIME);
                Trace("Display: Train Lesson\n");
                SelectNetErrLesson(1);
            }
        }
        
        if ((outputCount > 0) && outputToggleTimer.IsElapsed())
        { // It is time to toggle the monitored output neuron
            outputNeuron++;
            if (outputNeuron > outputCount)
            {
                outputNeuron = 1;
            }
            
            SelectPatternErrorViewerData(outputNeuron);
            outputToggleTimer.Start(TOGGLE_MONITORED_OUTPUT_INTERVAL);
        }
        
        // Leave some processing time for MemBrain itself.
        Sleep(100);
    }
    
    // Stop the teaching
    StopTeaching();
}


// Record the validation lesson and save it to file
void RecordValidationLesson()
{
    // Disable lesson editor update in order to improve performance
	EnableLessonEditorUpdate(false);
	
    SetLessonCount(3);
    SelectLesson(2);
 
    // Adjust the view
    ViewSetting(UPDATE_THINK, false);
    
    // Now prepare lesson #3 to record the output activation of the net
    SetRecordingType(RT_ACT);                // Record activation values of output neurons
    StartRecording(3);                       // Enable recording of data to lesson #2
    ShowLessonEditor(false);                 // Hide the lesson editor
    SetThinkSpeed(0);                        // Maximum think speed (0 ms wait between patterns)
    ThinkLesson();                           // Think on every pattern in lesson #1 (and record to #2)
    SleepExec();                             // Wait until 'Think on Lesson' is finished
    StopRecording();                         // Deactivate recording again
    
    // Append the original (target value) output columns to the recorded validation lesson
    AppendOutputTargetColumns(2, 3);
    
    // Now export the recorded lesson #3 to csv
    SelectLesson(3);
    ExportLessonRaw(sResultFileName);
    
    ViewSetting(UPDATE_THINK, true);// Re-enable update view on think steps
    string message = "The results from the net '" + sNetName + "' based on '" + sValidateLessonName + "' have been captured \n"
                     "in lesson #3 and also have been exported to '" + sResultFileName + "'";
                     
    MessageBox(message);
    
    EnableLessonEditorUpdate(true);	
}


// Append the original (target value) output columns to the recorded validation lesson
void AppendOutputTargetColumns(uint valiLesson, uint recordedLesson)
{
	SelectLesson(valiLesson);
	uint addedCount = GetLessonOutputCount();
	uint patternCount = GetLessonSize();
	SelectLesson(recordedLesson);
	uint oldOutCount = GetLessonOutputCount();
	SetLessonOutputCount(oldOutCount + addedCount);
	double act;
	string name;
	
	// Copy column names + name extention
	for (uint col = 1; col <= addedCount; col++)
	{  
		SelectLesson(valiLesson);
		GetLessonOutputName(col, name);
		SelectLesson(recordedLesson);
		SetLessonOutputName(oldOutCount + col, name + " " + sTargetColumnExtension);
	}
	
	// Copy data
	for (uint pat = 1; pat <= patternCount; pat++)
	{
		for (uint col = 1; col <= addedCount; col++)
		{  
			SelectLesson(valiLesson);
			SelectPattern(pat);
			GetPatternOutput(col, act);
			SelectLesson(recordedLesson);
			SelectPattern(pat);
			SetPatternOutput(oldOutCount + col, act);
		}
	}
}


// Reserved function name: 
// Called by MemBrain in case the script is aborted
void OnAbortScript()
{
    StopTeaching();
    StopThink();
    ViewSetting(UPDATE_THINK, true);// Re-enable update view on think steps
}
